require 'net/https'
require 'json'
require 'cgi'
require "net/http/post/multipart"

module Datarobot
  module AiApi
    class UnauthorizedError < StandardError; end;
    class NotFoundError < StandardError; end;
    class ApiError < StandardError; end;

    @@api_key = ''
    @@base_url = 'https://developers.datarobot.com'

    def self.api_key; @@api_key end
    def self.api_key= v; @@api_key = v end

    # Sets global API key
    # @param [String] api_key The api key to use for authentication
    # @return [Datarobot::AiApi]
    def self.configure!(api_key: '', base_url: nil)
      @@api_key = api_key
      @@base_url = base_url if base_url
      self
    end

    # Checks the `ping` endpoint. Useful for debugging
    # @return [Hash]
    def self.ping
      request_endpoint('/aiapi/ping')
    end

    # Checks the `ping/me` endpoint. Useful for debugging authentication
    # @return [Hash]
    def self.ping_me
      request_endpoint('/aiapi/ping/me')
    end

    # Adds default headers to a net http request
    # @param [Net::HTTP::*] request The net/http request to add the headers to
    # @param [Hash] additional_headers Will overwrite defaults 
    # @return [Net::Http::*] returns an aribitrary request type (GET/POST/etc.)
    def self.add_headers(request, additional_headers={})
      request["User-Agent"] = "datarobot-ai/ruby"
      request["Content-Type"] = "application/json"
      request["Authorization"] = "Bearer #{@@api_key}"
      additional_headers.each do |k, v|
        request[k] = v
      end
      request
    end

    # Make a request with https. This is a helper method to avoid having to
    # write net/http boilerplate
    # @param [String] uri The uri to make the request to
    # @param [Hash] params Additoinal params to add to the request. Will
    # overwrite params in the query string of the given uri
    #
    # @yield [uri, http] Yields the https request with the parsed uri and http object
    def self.with_https(uri, params={})
      uri = URI.parse(uri)
      existing_params = CGI::parse(uri.query.to_s)
      uri.query = URI.encode_www_form(existing_params.merge params)
      Net::HTTP.start(uri.host, uri.port, :use_ssl => uri.scheme == 'https') do |http|
        yield uri, http
      end
    end

    # Gets a bare uri. Useful for requests with `result` objects
    # @param [String] bare_uri
    # @param [Block] block passed to `handle_response`
    def self.get(bare_uri, &block)
      with_https(bare_uri) do |uri, http|
        request = Net::HTTP::Get.new(uri)
        request = add_headers(request)

        response = http.request(request)
        handle_response(response, &block)
      end
    end

    # Converts a file into an octet stream, and uploads it to given endpoint
    # @param [String] endpoint The endpoint to upload the file to
    # @param [String] file_path Path to local file for upload
    # @return Delegates return values to `handle_response`
    def self.upload_to_endpoint(endpoint, file_path, &block)
      uri = "#{@@base_url}#{endpoint}"
      with_https(uri) do |uri, http|
        request = Net::HTTP::Post::Multipart.new uri.request_uri, "file" => UploadIO.new(file_path, "application/octet-stream")
        request["User-Agent"] = "datarobot-ai/ruby"
        request["Authorization"] = "Bearer #{@@api_key}"
        response = http.request(request)
        handle_response(response, &block)
      end
    end

    # Common error handling and response parsing
    # @yield [data] Yields the parsed json response if there is one
    # @return nil if body is empty
    # @return [Hash] data Parsed json response if no block is provided
    def self.handle_response(response, &block)
      if response.code.to_i < 300
        return nil if response.body.to_s.empty?
        data = JSON.parse(response.body)
        if block_given?
          yield data
        else
          data
        end
      else
        determine_error(response)
      end
    end

    # Makes a request to the given endpoint
    # @param [String] endpoint The url path to the endpoint
    # @param [String] method The HTTP method to make the request with
    # @param [Hash] body JSON body. Only used in put and post requests
    # @param [Hash] headers Additional headers to add to request
    # @param [Hash] params Additoinal url parameters to add
    # @return Delegates return values to `handle_response`
    def self.request_endpoint(endpoint, method: 'get', body: {}, headers: {}, params: {}, &block)
      uri_str = "#{@@base_url}#{endpoint}"
      with_https(uri_str, params) do |uri, http|
        case method.downcase
        when 'put'
          request = Net::HTTP::Put.new(uri.request_uri)
          request.body = JSON.dump(body)
        when 'delete'
          request = Net::HTTP::Delete.new(uri.request_uri)
        when 'post'
          request = Net::HTTP::Post.new(uri.request_uri)
          request.body = JSON.dump(body)
        else
          request = Net::HTTP::Get.new(uri)
        end
        request = add_headers(request, headers)

        response = http.request(request)
        handle_response(response, &block)
      end
    end

    # Uses request data to determin error classes
    # Only raises values. Will never return
    def self.determine_error(response_obj)
      parsed = JSON.parse(response_obj.body)
      msg = parsed["error"] || parsed["message"]
      case response_obj.code
      when "403"
        raise Datarobot::AiApi::UnauthorizedError, msg
      when "404"
        raise Datarobot::AiApi::NotFoundError, msg
      else
        raise Datarobot::AiApi::ApiError, msg
      end
    end

    # Given an arbitrary url, return the correct AiApi object
    # @param [String] url
    # @return [Datarobot::AiApi::*]
    def self.determine_object(url)
      case url
      when /\/ais/
        Datarobot::AiApi::AI
      when /\/datasets/
        Datarobot::AiApi::Dataset
      when /\/learningSessions/
        Datarobot::AiApi::LearningSession
      when /\/outputs/
        Datarobot::AiApi::Output
      when /\/status/
        Datarobot::AiApi::Task
      when /\/deployment/
        Datarobot::AiApi::Deployment
      end
    end
  end
end

require 'datarobot/ai_api/page'
require 'datarobot/ai_api/refreshable'
require 'datarobot/ai_api/evaluation'
require 'datarobot/ai_api/deployment'
require 'datarobot/ai_api/ai'
require 'datarobot/ai_api/dataset'
require 'datarobot/ai_api/task'
require 'datarobot/ai_api/learning_session'
require 'datarobot/ai_api/output'
require 'datarobot/ai_api/prediction'