module Datarobot
  module AiApi

    # This is the primary model in the AI API. This is the thing that learns
    # about the data that you give it
    class AI
      include Datarobot::AiApi::Refreshable
      attr_reader :output_count, :dataset_count, :learning_session_count, :id
      attr_accessor :name,

      # Retrieves all AIs as a paginated resource.
      #
      # @param [Integer] limit Number of AIs per page
      # @param [Integer] offset How many AIs to skip before this page
      #
      # @return [Datarobot::AiApi::Page(Datarobot::AiApi::AI)]
      def self.all(limit: 50, offset: 0)
        Datarobot::AiApi.request_endpoint('/aiapi/ais/', params: {limit: limit, offset: offset}) do |data|
          Datarobot::AiApi::Page.new(self, data)
        end
      end

      # Retrieves an AI given an ID.
      #
      # @param [String] id The ID of the AI to retrieve
      #
      # @return [Datarobot::AiApi::AI]
      def self.find(id, &block)
        raise Datarobot::AiApi::NotFoundError, "Cannot find AI with id: nil" if id.nil?
        Datarobot::AiApi.request_endpoint("/aiapi/ais/#{id}") do |data|
          if block_given?
            yield data
          else
            self.new(data)
          end
        end
      end

      # Creates an AI
      #
      # @param [String] name The ID of the AI to retrieve
      #
      # @return [Datarobot::AiApi::AI]
      def self.create(name: 'New AI')
        Datarobot::AiApi.request_endpoint('/aiapi/ais/', method: 'post', body: { name: name }) do |data|
          ai_data = Datarobot::AiApi.get(data["links"]["result"])
          new(ai_data)
        end
      end

      # Deletes an AI. Returns `nil` if the action was successful. Will raise
      # an error if the action was unsuccessful
      #
      # @param [String] id The ID of the AI to delete
      #
      # @return [nil]
      def self.delete(id)
        Datarobot::AiApi.request_endpoint("/aiapi/ais/#{id}", method: "delete")
      end

      # Given a parsed response body from the API, will create a new AI object
      def initialize(options = {})
        # Suppresses warnings about uninitialized variables
        @id = nil
        @name = nil
        @dataset_count = nil
        @output_count = nil
        @learning_session_count = nil

        set_from_options(options)
        @outputs = nil
      end

      # Takes a response body from the API. Will set all AI attributes from the
      # response body
      #
      # @param [Hash] options A parsed response body
      # @return [void]
      def set_from_options(options = {})
        # one-liner replacement for `stringify_keys`
        options = options.collect{|k,v| [k.to_s, v]}.to_h

        @output_count = options.dig("outputCount") || @output_count
        @dataset_count = options.dig("datasetCount") || @dataset_count
        @learning_session_count = options.dig("learningSessionCount") || @learning_session_count
        @name = options.dig("name") || @name
        @id = options.dig("id") || @id
      end

      # Lists all the outputs associated with the AI as a paginated resource
      #
      # @return [Datarobot::AiApi::Page(Datarobot::AiApi::Output)]
      def outputs
        Datarobot::AiApi.request_endpoint("/aiapi/ais/#{@id}/outputs") do |data|
          data["aiId"] = @id
          Datarobot::AiApi::Page.new(Datarobot::AiApi::Output, data)
        end
      end

      # Lists all the datasets associated with the AI as a paginated resource
      #
      # @return [Datarobot::AiApi::Page(Datarobot::AiApi::Dataset)]
      def datasets
        Datarobot::AiApi.request_endpoint("/aiapi/datasets?aiId=#{@id}") do |data|
          data["aiId"] = @id
          Datarobot::AiApi::Page.new(Datarobot::AiApi::Dataset, data)
        end
      end

      # Finds an output by name
      #
      # @param [String] name The name of the output associated with this AI
      # that you want to find
      # @return [Datarobot::AiApi::Output]
      def find_output(name)
        # TODO: Update this when AI API supports CGI escaped URIs
        #
        # This code was lifted from source of erb method:
        # https://apidock.com/ruby/v2_6_3/ERB/Util/url_encode
        #
        # URI.escape is deprecated because it does not encode url control
        # characters, thus making it a potential security issue and CGI.escape
        # substitutes whitespace with a + which is correct, but unsupported by
        # the AI API
        encoded_name = name.gsub(/[^a-zA-Z0-9_\-.~]/) { |m|
          sprintf("%%%02X", m.unpack1("C"))
        }

        Datarobot::AiApi.request_endpoint("/aiapi/ais/#{@id}/outputs/#{encoded_name}") do |data|
          data["aiId"] = @id
          Datarobot::AiApi::Output.new(data)
        end
      end

      # Adds a dataset to the AI. Raises an error on failure. Returns true on
      # success
      #
      # @param [String] dataset_id The ID of the dataset to add
      # @return [Bool] true
      def add_dataset(dataset_id)
        Datarobot::AiApi.request_endpoint("/aiapi/ais/#{@id}/datasets", method: 'post', body: { datasetId: dataset_id })
        true
      end

      # Adds a learning session to the AI. Raises an error on failure. Returns
      # true on success
      #
      # @param [String] session_id The ID of the learning session to add
      # @return [Bool] true
      def add_learning_session(session_id)
        Datarobot::AiApi.request_endpoint("/aiapi/ais/#{@id}/learningSessions", method: 'post', body: { learningSessionId: session_id })
        true
      end

      # Trains an AI on the given target with the data in the file at the given
      # path. Returns true if successful. Raises an error otherwise
      #
      # @param [String] target The target to train the AI on
      # @return [Bool] true
      def train(target, file_path)
        dataset = Datarobot::AiApi::Dataset.create_from_file(file_path)
        add_dataset(dataset.id)
        learning_session = Datarobot::AiApi::LearningSession.create(dataset_id: dataset.id, target: target)
        add_learning_session(learning_session.id)
        Datarobot::AiApi::Output.create(ai_id: @id, learning_session_id: learning_session.id, output_name: "#{@name} output")
        true
      end

      # Predicts a target feature given a data hash
      #
      # @param target The target feature to predict
      # @param data The remaining features used to predict target feature
      def predict(target, data={})
        output = outputs.find { |o| o.target == target }

        raise NotFoundError, "No output with target #{target.inspect} found AI with ID #{@id.inspect}" if output.nil?

        deployment_id = output.source["deploymentId"]
        key = output.source["datarobot-key"]
        Datarobot::AiApi.request_endpoint(
          "/predApi/v1.0/deployments/#{deployment_id}/predictions/",
          method: 'post',
          body: [ data ],
          headers: {"datarobot-key" => key}
        ) do |data|
          Datarobot::AiApi::Page.new(Datarobot::AiApi::Prediction, data)
        end
      end
    end
  end
end